 /* Reduccion  por el metodo de Quine Mc Klusqui */

#ifndef _QUINE_H_
#define _QUINE_H_

#include <stdio.h>
#include <string.h>

class Quine
{

	// Variables Globales
	int n;

	struct nodo {
		char *cad;
		int usada:1;
		struct nodo *sig;
		nodo() { cad = NULL; usada = 0; sig = NULL; }
		nodo(char *cc) { cad = cc; usada = 0; sig = NULL; }
	} *Final, *Aux, **Arr1, *iterator;
	
	// Funciones
	char *Simplificar(char *cad1, char *cad2)
    {
	     int dif=0;
	     char *res = new char[n+1];
	     res[n] = '\0';
    	for (int i=0;(i < n) && (dif < 2);i++)
        {
    		if ( cad1[i] != cad2[i] )
            {
    		 res[i] = 'x';
    		 dif++;
    		}
            else res[i] = cad1[i];
    	}
    	if ( dif < 2 ) return (res);
    	else
        {
            delete[] res;
    		return (NULL);    
    	}
	}
	
	int ContarUnos(char *cad)
    {
    	int cnt = 0;
    	for ( int i=0; i < n ; i++ ) 
    		cnt += (cad[i] == '1');
    	return cnt;
	}
	
	struct nodo *EliminarLista(struct nodo* &pt)
    {
		if (pt)
        {
    		EliminarLista(pt->sig);
    		delete[] pt->cad;
    		delete pt;
		}
		return (NULL);
	}
	
	struct nodo **EliminarArreglo(struct nodo **Arr,int size)
    {
    	for (int i = 0; i < size; i++)
        {
    		EliminarLista(Arr[i]);
    		Arr[i] = NULL;
    	}
    	delete[] Arr;
    	return NULL;
	}
	
	void InsertarNodo(struct nodo* nuevo, struct nodo **Arr)
    {
		if (nuevo)
        {
    		if (*Arr)
            {
    			nuevo->sig = *Arr;
    		}
            else nuevo->sig = NULL;
    		*Arr = nuevo;
		}
	}
	
	int Existe(struct nodo *cad1, struct nodo *lista)
	// Busca en Lista la existencia de Cad. Retorna 1 si existe, 0 en caso contrario.
	{
    	if (cad1)
        {
    		while (lista)
            {
        		if ( !strcmp(lista->cad,cad1->cad) ) return 1; // Existe.
        		else lista = lista->sig;
    		}
    	}
    	return 0;
	}
	
	int VaciarArreglo_BuscarSolos(struct nodo **Arr, int size)
	/* Esta funcion se encarga de buscar las evaluaciones  que no fueron "marcadas" y las
		agrega a la lista Final" a medida que va vaciando el arreglo y lo deja todo en NULL.
	Retorna 0 si ninguna evaluacion se marco; 1 en caso contrario. 			*/
	{
    	struct nodo *aux1 = NULL, *aux2 = NULL;
    	int fi = 0; // Se marcaron evaluaciones ??
    	for (int i = 0; i < size; i++)
        {
    		if ( aux1 = Arr[i] )
            {
    			while ( aux1 )
                {
    				if ( !aux1->usada && !Existe(aux1, Final) ) // NO marcada Y NO existe en Final.
    				{
        				aux2 = aux1->sig;
        				InsertarNodo(aux1, &Final);
        				aux1 = aux2;
    				}
                    else
                    {
        				aux2 = aux1->sig;
        				delete[] aux1->cad;
        				delete aux1;
        				aux1 = aux2;
        				fi = 1;
    				}
    			}
    			Arr[i] = NULL;
    		}
    	}
    	return fi;
	}
	
	struct nodo *ListaSimplificada(struct nodo *l1, struct nodo *l2)
	{
		struct nodo *Res = NULL, *naux = NULL;
		if ( l1 && l2 )
        {
    		char *cadsimp;
    		struct nodo *aux;
    		while ( l1 )
            {
    			aux = l2;
    			while ( aux )
                {
        			if ( cadsimp = Simplificar(l1->cad, aux->cad) )
                    {
        				naux = new struct nodo(cadsimp);
        				l1->usada = 1;
        				aux->usada = 1;
        				InsertarNodo(naux, &Res);
        			}
        			aux = aux->sig;
    			}
    			l1 = l1->sig;
    		}
		}
		return Res;
	}
/*	
	void MostrarArreglo(struct nodo **Arr, int size)
    {
		for (int i = 0; i < size; i++)
        {
    		if (Arr[i])
            {
    			struct nodo *aux = Arr[i];
    			while (aux)
                {
        			printf("%s\n",aux->cad);
        			aux = aux->sig;
    			}
    			printf("\n"); // Para separar los grupos de distinta cantidad de unos.
    		}
		}
	}
	
	void MostrarLista(struct nodo *Lis)
    {
			while (Lis)
            {
    			printf("%s\n",Lis->cad);
    			Lis = Lis->sig;
			}
			printf("\n");
	}
*/	
	int CadOK( const char *in )
	// Verifica que IN solo este compuesta por ceros y/o unos.
	// Si es asi devuelve la longitud de la cadena; -1 en caso contrario. 
	{
    	int i = 0;
    	if (in)
        {
    		while ( in[i] && ( (in[i] == '1') || (in[i] == '0') ) ) i++;
    		if (in[i]) return(-1);
    	}
    	return i;
	}


   public:
    Quine() //Constructor
    {
           n = 0;
           Final = Aux = iterator = NULL;
           Arr1=NULL;
    }
    
    ~Quine() // Destructor
    {
            EliminarLista( Final );
	    if ( Arr1 )
		Arr1 = EliminarArreglo(Arr1,n+1);
    }
	// MINIMIZE toma el arreglo con las evaluaciones ordenadas segun la cantidad de 1 que poseen 
	// y almacena en FINAL los datos simplificados.
	void minimize ( struct nodo **Arr )
	{
		int seguir;
		struct nodo **ArrAux = new struct nodo*[n+1];
		for (int i = 0; i <= n; i++) ArrAux[i] = NULL;
		do {
			for (int i = 0; i < n; i++) {
			ArrAux[i] = ListaSimplificada(Arr[i],Arr[i+1]);
			}
			if ( seguir = VaciarArreglo_BuscarSolos(Arr, n+1) ) {
			for (int i = 0; i < n; i++)
				Arr[i] = ListaSimplificada(ArrAux[i],ArrAux[i+1]);
			seguir = VaciarArreglo_BuscarSolos(ArrAux, n+1);
			}
		} while (seguir);
		ArrAux = EliminarArreglo(ArrAux,n+1);
		
		iterator = Final; // Para poder recorrer toda la lista.
	}
	
	// Realiza la Simplificacion !!! llamar este metodo desde la interface
	void doQuine ()
	{
	   minimize ( Arr1 );
	}

	// Agrega una valuacion a la lista de valuaciones verdaderas
	bool add ( const char *cadena )
	{
		if ( (CadOK(cadena) == n ) /*&& (*cadena != '*')*/ )
		{
			char *temp = new char[n+1];
			strcpy ( temp, cadena );
			Aux = new struct nodo(temp);
			InsertarNodo(Aux,&Arr1[ContarUnos(temp)]);
		} else {
			printf("		Evaluacion incorrecta: %s\n",cadena);
			return false;
		}
		return true;
	}

	// Setea la cantidad de variables de la funcion
	void setSize ( int size )
	{
                Final = EliminarLista( Final );
	        if ( Arr1 )
		    Arr1 = EliminarArreglo(Arr1,n+1);
		n = size;
		Arr1 = new struct nodo*[n+1];
		for (int i = 0; i <= n; i++)
			Arr1[i] = NULL;
	}
	
	const char *resultado ()
	// Retorna uno a uno los valores de la lista FINAL. Indica que termino el recorrido al retornar NULL
	{
               if ( iterator )
               {
                  char *restmp = iterator->cad;
                  iterator = iterator->sig;
                  return restmp;                  
               }
               else
                   return NULL;
    }

};

/*******************************************
	int main() {
	fflush(stdin);
	printf("\n\nPrograma de Reduccion de expresiones Booleanas\n");
	printf("Ingrese el Numero de VARIABLES:");
			scanf("%d",&n);
	printf("\nIngrese las entradas de funcion para las cuales la evaluacion da como resultado '1'.\n");
	printf("Separados por espacio o enter (por ejemplo: 1110011 1101000 para n=7); Finaliza con '*' :\n");
	char *temp;
			Arr1 = new struct nodo*[n+1];
			for (int i = 0; i <= n; i++) Arr1[i] = NULL;
	fflush(stdin);
	int Cont = 0;
	do {
	temp = new char[2*n];
	scanf("%s",temp);
			if ( (CadOK(temp) == n ) && (*temp != '*') ) {
				Cont++;
				Aux = new struct nodo(temp);
				InsertarNodo(Aux,&Arr1[ContarUnos(temp)]);
			} else if (*temp != '*') printf("		Evaluacion incorrecta: %s\n",temp);
	} while (temp[0] != '*');
	printf("\n	Se ingresaron %3d evaluaciones.\n",Cont); // A esta altura ya tengo las variables cargadas
	
			QUINE(Arr1);
	
	printf("\nResultado de la simplificacion (las 'x' significan que la variable no aparece):\n");
	MostrarLista(Final);
	EliminarLista(Final);
	//    MostrarArreglo(Arr1,n+1);
			Arr1 = EliminarArreglo(Arr1,n+1);
	printf("\n Desarrollado por Leonardo Roman - Center-Tech 2004 -- Tel. (02983) 15640199.\n");
	printf(" Programa Freeware - Prohibida su venta.\n\n");
	return 0;
	}
********************************/
#endif  //_QUINE_H_ 
